/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2011 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA
 * 02110-1335, USA.
 */

#ifndef __CSM_ICON_NAMES_H__
#define __CSM_ICON_NAMES_H__

#define CSM_ICON_COMPUTER_FAIL       "computer-fail"
#define CSM_ICON_INHIBITOR_DEFAULT   "gnome-windows"
#define CSM_ICON_LOGOUT              "system-log-out"
#define CSM_ICON_SHUTDOWN            "system-shutdown"
#define CSM_ICON_XSMP_DEFAULT        "system-run"

#endif /*__CSM_ICON_NAMES_H__ */
